# File: tello_rc.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 29 Jun 2023
# Purpose:
#   A class for handling a multi-processed tello drone using keyboard controls.
# Notes:


import multiprocessing as mp
import queue
from time import perf_counter
from math import log1p
import pygame as pg
import sys
from threading import Thread

from .tello_remote import tello_remote_loop
from .tello_state import tello_state_loop
from .tello_video import tello_video_loop


class TelloRC:
    # Precond:
    #   The computer creating the TelloDrone instance is connected to the Tello's Wi-Fi.
    #
    # Postcond:
    #   Sets up a connection with the Tello Drone.
    def __init__(self):
        # Setup command process
        self.rcQ = mp.Queue()
        self.rc_confQ = mp.Queue()
        self.rc_process = mp.Process(target=tello_remote_loop, args=(self.rcQ, self.rc_confQ))
        self.rc_process.daemon = True
        
        # Setup state process
        self.state_haltQ = mp.Queue()
        self.state_recQ = mp.Queue()
        self.state_process = mp.Process(target=tello_state_loop, args=(self.state_haltQ, self.state_recQ))
        self.state_process.daemon = True
        self.state_thread = Thread(target=self.__state_thread)
        self.state_thread.daemon = True
        
        # Setup video process
        self.video_haltQ = mp.Queue()
        self.video_recQ = mp.Queue()
        self.video_process = mp.Process(target=tello_video_loop, args=(self.video_haltQ, self.video_recQ))
        self.video_process.daemon = True
        self.video_thread = Thread(target=self.__video_thread)
        self.video_thread.daemon = True
        
        # Internal variables
        self.current_rc = [0, 0, 0, 0]
        self.last_state = None
        self.last_frame = None
        self.running = False
        self.vel_timing = 10

    # ==========================
    #   MANAGEMENT METHODS
    # ==========================

    # Precond:
    #   None.
    #
    # Postcond:
    #   Starts the TelloRC object.
    #   Connects to the Tello.
    #   Connects to all streams and begins all processes.
    #   Returns True if all processes have been started.
    def start(self):
        self.running = True
        self.rc_process.start()
        self.state_thread.start()
        self.state_process.start()
        self.video_thread.start()
        self.video_process.start()
        return True
    
    def control(self):
        # Setup
        key_holds = {'w': 0, 's': 0, 'd': 0, 'a': 0, 'q': 0, 'e': 0, 'r': 0, 'f': 0}
        poll_timer = perf_counter()
        poll_delta = 1/30
        # Main Loop
        control_running = True
        pg.init()
        while control_running:
            delta = perf_counter() - poll_timer
            if delta >= poll_delta:
                # Check for events
                command_sent = False
                for event in pg.event.get(pg.KEYDOWN):
                    if event.key == pg.K_t:
                        command_sent = True
                        if not self.takeoff():
                            print("Problem with takeoff!")
                    elif event.key == pg.K_l:
                        command_sent = True
                        if not self.land():
                            print("Problem with landing!")
                    elif event.key == pg.K_ESCAPE:
                        command_sent = True
                        if not self.emergency():
                            print("Problem with emergency shutdown!")
                    elif event.key == pg.K_BACKSPACE:
                        control_running = False
                    elif event.key == pg.K_DELETE:
                        self.rcQ.put((0, 0, 0, 0))
                    elif event.key == pg.K_UP:
                        command_sent = True
                        if not self.flip_forward():
                            print("Problem with forward flip!")
                    elif event.key == pg.K_DOWN:
                        command_sent = True
                        if not self.flip_backward():
                            print("Problem with backward flip!")
                    elif event.key == pg.K_RIGHT:
                        command_sent = True
                        if not self.flip_right():
                            print("Problem with right flip!")
                    elif event.key == pg.K_LEFT:
                        command_sent = True
                        if not self.flip_left():
                            print("Problem with left flip!")
                if command_sent:
                    poll_timer = perf_counter()
                    continue
                # Deal with held keys
                key_state = pg.key.get_pressed()
                for key in key_holds:
                    if key_state[pg.key.key_code(key)]:
                        key_holds[key] += delta
                    else:
                        key_holds[key] -= delta
                    key_holds[key] = max(0, min(self.vel_timing, key_holds[key]))
                y = self.__vel_curve(key_holds['w']) - self.__vel_curve(key_holds['s'])
                x = self.__vel_curve(key_holds['d']) - self.__vel_curve(key_holds['a'])
                z = self.__vel_curve(key_holds['r']) - self.__vel_curve(key_holds['f'])
                rot = self.__vel_curve(key_holds['e']) - self.__vel_curve(key_holds['q'])
                if self.rcQ.empty():
                    self.rcQ.put((int(x), int(y), int(z), int(rot)))
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Closes down communication with the drone and writes the log to a file.
    def close(self):
        self.running = False
        self.state_thread.join()
        self.video_thread.join()
        self.video_haltQ.put("halt")
        TelloRC.__clear_q(self.video_recQ)
        self.video_process.join()
        self.state_haltQ.put("halt")
        TelloRC.__clear_q(self.state_recQ)
        self.state_process.join()
        self.rcQ.put("halt")
        TelloRC.__clear_q(self.rc_confQ)
        self.rc_process.join()

    # ======================================
    # COMMAND METHODS
    # ======================================
    # Section Notes:
    #   All commands check to see if the drone has been connected and put into SDK mode before sending commands.

    # Precond:
    #   None.
    #
    # Postcond
    #   Sends the takeoff command.
    #   If a non-okay response is given returns False, otherwise returns true.
    def takeoff(self):
        self.rcQ.put("takeoff")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except queue.Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond
    #   Sends the land command.
    #   If a non-okay response is given returns False, otherwise returns true.
    def land(self):
        self.rcQ.put("land")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except queue.Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends the emergency command, in triplicate. Does not wait for response.
    def emergency(self):
        self.rcQ.put("emergency")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except queue.Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Adds flip l command to the command queue.
    def flip_left(self):
        self.rcQ.put("flip l")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except queue.Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Adds flip r command to the command queue.
    def flip_right(self):
        self.rcQ.put("flip r")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except queue.Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Adds flip f command to the command queue.
    def flip_forward(self):
        self.rcQ.put("flip f")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except queue.Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Adds flip b command to the command queue.
    def flip_backward(self):
        self.rcQ.put("flip b")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except queue.Empty:
            return False
        return True

    # ======================================
    # Info METHODS
    # ======================================

    # Precond:
    #   None.
    #
    # Postcond:
    #   Returns the last frame taken by the Tello.
    #   Returns None if the stream is off.
    def get_frame(self):
        return self.last_frame

    # Precond:
    #   None.
    #
    # Postcond:
    #   Returns the last state received from the Tello as a dictionary.
    def get_state(self):
        return self.last_state

    # ======================================
    # PRIVATE METHODS
    # ======================================

    # Precond:
    #   None.
    #
    # Postcond:
    #   Thread handling state extraction..
    def __state_thread(self):
        while self.running:
            self.last_state = self.state_recQ.get()

    # Precond:
    #   None.
    #
    # Postcond:
    #   Thread handling video extraction.
    def __video_thread(self):
        while self.running:
            self.last_frame = self.video_recQ.get()

    # Precond:
    #   q is a valid mp.Queue object.
    #
    # Postcond:
    #   Clears the q and closes it.
    def __clear_q(q: mp.Queue):
        try:
            while not q.empty():
                q.get_nowait()
        except queue.Empty:
            pass
        q.close()
    
    # Precond:
    #   t is a valid floating point value.
    #
    # Postcond:
    #   Returns the current velocity based on how long a key has  been pressed.
    def __vel_curve(self, t):
        return 100 * (log1p(t) / log1p(self.vel_timing))


# File: tello_state.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 26 Jun 2023
# Purpose:
#   A class and function for managing getting the state of a Tello in a separate process.
# Notes:
#   This is intended for use in a multiprocessing capacity.
#   Start by running tello_state_loop as the target of a Process object.
#
#   Two multiprocessing queues are used:
#       halt_q: Used by the parent to send commands to the state gatherer.
#               Commands should be sent as a single string.
#       conf_q: Used by the child to send the next frame gathered from the Tello.
#
#   Commands (case insensitive) accepted:
#       halt: Stops the process and closes the management object.

from threading import Thread
import multiprocessing as mp
import cv2
import queue


def tello_video_loop(halt_q: mp.Queue, frame_q: mp.Queue):
    # Create a management object.
    manager = TelloVideo()
    manager.start()
    running = True
    while running:
        try:
            halt = str(halt_q.get(False))
            if halt.lower() == 'halt':
                break
        except queue.Empty:
            pass
        if frame_q.empty():
            frame = manager.get()
            if frame is not None:
                frame_q.put(frame)
    try:
        while not frame_q.empty():
            frame_q.get_nowait()
    except queue.Empty:
        pass
    frame_q.close()
    manager.close()


# Class for handling receiving video frames from a Tello.
class TelloVideo:
    def __init__(self):
        # Running info
        self.last_frame = None
        self.frame_update = False

        # Connecting the video
        self.video_connect_str = 'udp://192.168.10.1:11111'
        self.video_stream = None
        self.video_thread = Thread(target=self.__receive)
        self.video_thread.daemon = True
        self.last_frame = None
        self.stream_active = True
        self.frame_width = 0
        self.frame_height = 0
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Starts the state receiving thread.
    def start(self):
        # Set up the video stream
        self.stream_active = True
        self.video_stream = cv2.VideoCapture(self.video_connect_str)
        self.video_stream.set(cv2.CAP_PROP_BUFFERSIZE, 2)
        self.frame_width = self.video_stream.get(cv2.CAP_PROP_FRAME_WIDTH)
        self.frame_height = self.video_stream.get(cv2.CAP_PROP_FRAME_HEIGHT)
        self.video_thread.start()
        self.frame_update = False
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Non-blocking function.
    #   Returns the most recent frame.
    #   If no new frame has been received then None is returned.
    def get(self):
        if self.frame_update:
            self.frame_update = False
            return self.last_frame
        return None
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Closes the threads and various channels.
    def close(self):
        self.stream_active = False
        self.video_thread.join()
        self.last_frame = None

    # Precond:
    #   None.
    #
    # Postcond:
    #   Receives video messages from the Tello.
    def __receive(self):
        while self.stream_active:
            ret, img = self.video_stream.read()
            if ret:
                self.last_frame = (img, self.frame_width, self.frame_height)
                self.frame_update = True
        self.video_stream.release()
        cv2.destroyAllWindows()


# File: tello_drive.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 27 Jun 2023
# Purpose:
#   A program for testing flying a tello drone from a ground station computer.
# Notes:

from .tello_drone import TelloDrone
from .tello_rc import TelloRC
from threading import Thread
from .tello_hud import TelloHud
from time import sleep


def drive_tello():
    drone = TelloRC()
    if not drone.start():
        print("Problem connecting.")
        return False
    hud = TelloHud(drone)
    hud.activate_hud()
    drone.control()
    hud.deactivate_hud()
    drone.close()

    
if __name__ == '__main__':
    drive_tello()
    
    


# File: tello_face_hud.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 29 Jun 2023
# Purpose:
# Notes:

from .tello_drone import TelloDrone
from .tello_rc import TelloRC
from .face_encoder import FaceEncoder
from time import perf_counter
import pygame as pg
from threading import Thread


class TelloFaceHud:
    def __init__(self, drone: TelloDrone | TelloRC, faces: FaceEncoder):
        self.drone = drone
        self.encoder = faces
        self.running = False
        self.hud_thread = Thread(target=self.__hud_stream)
        self.hud_thread.daemon = True
        
        #pg reqs
        self.font = pg.font.Font(pg.font.get_default_font(), 16)
    
    def activate_hud(self):
        if self.running:
            return
        self.running = True
        self.hud_thread.start()
    
    def deactivate_hud(self):
        if not self.running:
            if self.hud_thread.is_alive():
                self.hud_thread.join()
            return
        self.running = False
        self.hud_thread.join()
    
    def is_active(self):
        return self.running
    
    def __hud_stream(self):
        # Setup Pygame
        pg.display.init()
        screen = pg.display.set_mode((640, 480))
        pg.font.init()
        hud_font = pg.font.Font(pg.font.get_default_font(), 20)
        # Setup video loop basics
        frame_timer = perf_counter()
        frame_delta = 1 / 24
        while self.running:
            if (perf_counter() - frame_timer) > frame_delta:
                frame_timer = perf_counter()
                frame = self.drone.get_frame()
                if frame is not None:
                    frame = frame[0]
                    frame = pg.image.frombuffer(frame.tobytes(), frame.shape[1::-1], "BGR")
                    screen.blit(frame, (0, 0))
                    # Detect Faces
                    faces = self.encoder.detect_faces(frame)
                    for name, location in faces:
                        rect = TelloFaceHud.__convert_rect(location)
                        pg.draw.rect(frame, (0, 200, 0), rect, 1)
                        name_text = self.font.render(name, True, (0, 200, 0))
                        frame.blit(name_text, (rect.x, rect.y-name_text.get_height()-1))
                pg.display.flip()
                for event in pg.event.get(pg.QUIT):
                    self.running = False
        pg.display.quit()
    
    @staticmethod
    def __convert_rect(rect):
        top_left = rect[3], rect[0]
        width_height = rect[1]-rect[2], rect[2]-rect[0]
        return pg.Rect(top_left, width_height)


# File: tello_cmd.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 26 Jun 2023
# Purpose:
#   A class and function for managing sending commands to a Tello in a separate process.
# Notes:
#   This is intended for use in a multiprocessing capacity.
#   Start by running tello_command_loop as the target of a Process object.
#
#   Two multiprocessing queues are used:
#       cmd_q: Used by the parent to send commands to the Tello.
#              Commands should be sent as a single string.
#       conf_q: Used by the child to send confirmation of completion/failure to parent.
#               Messages in queue are (boolean, string) pairs.
#               boolean indicates success/failure, string give setails.
#
#   Commands (case insensitive) accepted:
#       halt: Stops the process and closes the management object.
#       takeoff: Makes the Tello takeoff.
#       land: Makes the Tello land.
#       up <dist>: Moves the Tello up to <dist> cm above ground level.
#       down <dist>: Moves the Tello down to <dist> cm above ground level.
#       left <dist>: Moves the Tello left <dist> cm.
#       right <dist>:  Moves the Tello right <dist> cm.
#       forward <dist>:  Moves the Tello forward <dist> cm.
#       backward <dist>:  Moves the Tello backward <dist> cm.
#       rotates [cw, ccw] <deg>: Rotates the Tello <deg> degrees [cw, ccw].
#       flip [f, b, l, r]: Flips the drone in a direction ([f, b, l, r]).
#       move <x> <y> <z> <spd>: Moves the Tello by the given (x, y, z) in cm at speed spd.
#       stream [on, off]: Turns the Tello video stream [on, off].
#       emergency: send command to shut down Tello (no confirmation).

import multiprocessing as mp
from socket import socket, AF_INET, SOCK_DGRAM
from threading import Thread
from time import perf_counter
from datetime import datetime
import queue
import os


def tello_command_loop(cmd_q: mp.Queue, conf_q: mp.Queue):
    # Create a management object
    manager = TelloCmd()
    res = manager.startup()
    conf_q.put((res, 'startup'))
    if not res:
        return
    running = True
    while running:
        try:
            cmd = str(cmd_q.get(False)).lower()
            cmd = cmd.split(' ')
            # Paring the send commands
            try:
                if cmd[0] == "halt":
                    running = False
                elif cmd[0] == "takeoff":
                    res = manager.takeoff()
                    conf_q.put((res, 'takeoff'))
                elif cmd[0] == "land":
                    res = manager.land()
                    conf_q.put((res, 'land'))
                elif cmd[0] == "up":
                    res = manager.up(int(cmd[1]))
                    conf_q.put((res, 'up'))
                elif cmd[0] == "down":
                    res = manager.down(int(cmd[1]))
                    conf_q.put((res, 'down'))
                elif cmd[0] == "left":
                    res = manager.left(int(cmd[1]))
                    conf_q.put((res, 'left'))
                elif cmd[0] == "right":
                    res = manager.right(int(cmd[1]))
                    conf_q.put((res, 'right'))
                elif cmd[0] == "forward":
                    res = manager.forward(int(cmd[1]))
                    conf_q.put((res, 'forward'))
                elif cmd[0] == "backward":
                    res = manager.backward(int(cmd[1]))
                    conf_q.put((res, 'backward'))
                elif cmd[0] == "rotate":
                    res = False
                    if cmd[1] == 'cw':
                        res = manager.rotate_cw(int(cmd[2]))
                    elif cmd[1] == 'ccw':
                        res = manager.rotate_ccw(int(cmd[2]))
                    conf_q.put((res, 'rotate ' + cmd[1]))
                elif cmd[0] == "flip":
                    res = False
                    if cmd[1] == 'f':
                        res = manager.flip_forward()
                    elif cmd[1] == 'b':
                        res = manager.flip_backward()
                    elif cmd[1] == 'l':
                        res = manager.flip_left()
                    elif cmd[1] == 'r':
                        res = manager.flip_right()
                elif cmd[0] == "move":
                    res = manager.move(int(cmd[1]),
                                       int(cmd[2]),
                                       int(cmd[3]),
                                       int(cmd[4]))
                    conf_q.put((res, 'move'))
                elif cmd[0] == "stream":
                    if cmd[1] == "on":
                        res = manager.stream_on()
                        conf_q.put((res, 'stream on'))
                    elif cmd[1] == "off":
                        res = manager.stream_on()
                        conf_q.put((res, 'stream off'))
                elif cmd[0] == "emergency":
                    manager.emergency()
            except ValueError:
                conf_q.put((False, 'Invalid Argument'))
        except queue.Empty:
            pass
    try:
        while not cmd_q.empty():
            cmd_q.get_nowait()
    except queue.Empty:
        pass
    cmd_q.close()
    manager.close()
    

# Class for handling sending Tello commands and logging data.
class TelloCmd:
    def __init__(self):
        # Addresses
        self.local_addr = ('', 8889)
        self.tello_addr = ('192.168.10.1', 8889)
    
        # Setup channels
        self.send_channel = socket(AF_INET, SOCK_DGRAM)
        self.send_channel.bind(self.local_addr)
    
        self.stop = True
        self.connected = False
    
        # Setup receiving thread
        self.receive_thread = Thread(target=self.__receive)
        self.receive_thread.daemon = True
    
        # Setup logs
        self.log = []
        self.MAX_TIME_OUT = 10  # measured in seconds
        
    def startup(self):
        self.stop = False
        self.receive_thread.start()
        if self.__connect(5):
            return True
        return False

    # ======================================
    # COMMAND METHODS
    # ======================================
    # Section Notes:
    #   All commands check to see if the drone has been connected and put into SDK mode before sending commands.

    # Precond:
    #   None.
    #
    # Postcond
    #   Sends the takeoff command.
    #   If a non-okay response is given returns False, otherwise returns true.
    def takeoff(self):
        if not self.connected:
            return False
        res = self.__send("takeoff")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond
    #   Sends the land command.
    #   If a non-okay response is given returns False, otherwise returns true.
    def land(self):
        if not self.connected:
            return False
        res = self.__send("land")
        return res is not None and res == 'ok'

    # Precond:
    #   val is an integer representing the amount to move
    #
    # Postcond:
    #   Moves the drone up to val centimeters.
    #   Returns False if the command could not be sent for any reason.
    def up(self, val):
        if not self.connected or val not in range(20, 501):
            return False
        res = self.__send("up " + str(val))
        return res is not None and res == 'ok'

    # Precond:
    #   val is an integer representing the amount to move
    #
    # Postcond:
    #   Moves the drone down to val centimeters.
    #   Returns False if the command could not be sent for any reason.
    def down(self, val):
        if not self.connected or val not in range(20, 501):
            return False
        res = self.__send("down " + str(val))
        return res is not None and res == 'ok'

    # Precond:
    #   val is an integer representing the amount to move
    #
    # Postcond:
    #   Moves the drone right val centimeters.
    #   Returns False if the command could not be sent for any reason.
    def right(self, val):
        if not self.connected or val not in range(20, 501):
            return False
        res = self.__send("right " + str(val))
        return res is not None and res == 'ok'

    # Precond:
    #   val is an integer representing the amount to move
    #
    # Postcond:
    #   Moves the drone left val centimeters.
    #   Returns False if the command could not be sent for any reason.
    def left(self, val):
        if not self.connected or val not in range(20, 501):
            return False
        res = self.__send("left " + str(val))
        return res is not None and res == 'ok'

    # Precond:
    #   val is an integer representing the amount to move
    #
    # Postcond:
    #   Moves the drone forward val centimeters.
    #   Returns False if the command could not be sent for any reason.
    def forward(self, val):
        if not self.connected or val not in range(20, 501):
            return False
        res = self.__send("forward " + str(val))
        return res is not None and res == 'ok'

    # Precond:
    #   val is an integer representing the amount to move
    #
    # Postcond:
    #   Moves the drone backward val centimeters.
    #   Returns False if the command could not be sent for any reason.
    def backward(self, val):
        if not self.connected or val not in range(20, 501):
            return False
        res = self.__send("back " + str(val))
        return res is not None and res == 'ok'

    # Precond:
    #   val is an integer representing the amount to move
    #
    # Postcond:
    #   Rotates the drone clockwise val degrees.
    #   Returns False if the command could not be sent for any reason.
    def rotate_cw(self, val):
        if not self.connected or val not in range(1, 361):
            return False
        res = self.__send("cw " + str(val))
        return res is not None and res == 'ok'

    # Precond:
    #   val is an integer representing the amount to move
    #
    # Postcond:
    #   Rotates the drone counterclockwise val degrees.
    #   Returns False if the command could not be sent for any reason.
    def rotate_ccw(self, val):
        if not self.connected or val not in range(1, 361):
            return False
        res = self.__send("ccw " + str(val))
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Flips the drone left.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def flip_left(self):
        if not self.connected:
            return False
        res = self.__send("flip l")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Flips the drone right.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def flip_right(self):
        if not self.connected:
            return False
        res = self.__send("flip r")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Flips the drone forward.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def flip_forward(self):
        if not self.connected:
            return False
        res = self.__send("flip f")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Flips the drone backward.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def flip_backward(self):
        if not self.connected:
            return False
        res = self.__send("flip b")
        return res is not None and res == 'ok'

    # Precond:
    #   x the amount to move in the x-axis.
    #   y the amount to move in the y-axis.
    #   z the amount to move in the z-axis.
    #   spd is the speed of movement
    #
    # Postcond:
    #   Moves the drone by the given x, y, and z values.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def move(self, x, y, z, spd):
        if not self.connected or not (20 < max(abs(x), abs(y), abs(z)) < 500) or spd not in range(10, 101):
            return False
        coord_str = ' '.join([str(x), str(y), str(z)])
        res = self.__send("go " + coord_str)
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends the emergency command, in triplicate. Does not wait for response.
    def emergency(self):
        for _ in range(3):
            self.__send_nowait("emergency")
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends a command to the Tello turning its video stream on.
    def stream_on(self):
        res = self.__send("streamon")
        return res is not None and res == 'ok'
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends a command to the Tello turing its video stream off.
    def stream_off(self):
        res = self.__send("streamoff")
        return res is not None and res =='ok'
    
    # Precond:
    #   fldr is a string containing the path to the folder to place the log file.
    #
    # Postcond:
    #   Closes down communication with the drone and writes the log to a file.
    def close(self, fldr: str = "logs"):
        self.connected = False
        self.stop = True
        self.send_channel.close()
        self.receive_thread.join()
        t = datetime.now()
        log_name = os.path.join(fldr, t.strftime("%Y-%m-%d_%H-%M-%S") + '-cmd.log')
        if not os.path.exists(fldr):
            os.mkdir(fldr)
        with open(log_name, 'w') as fout:
            count = 0
            for entry in self.log:
                print("Message[" + str(count) + "]:", entry[0], file=fout)
                print("Response[" + str(count) + "]:", entry[1], file=fout)
                count += 1

    # ======================================
    # PRIVATE METHODS
    # ======================================

    # Precond:
    #   attempts is the number of times to try and connect.
    #
    # Postcond:
    #   Checks connection to the drone by sending a message to
    #     switch the drone into SDK mode.
    #   Returns true if the connection was made.
    #   Returns false if there was a problem connecting and attempts were
    #       exceeded.
    def __connect(self, attempts=5):
        for _ in range(attempts):
            res = self.__send("command")
            if res is not None and res == 'ok':
                self.connected = True
                return True
        return False

    # Precond:
    #   mess is a string containing the message to send.
    #
    # Postcond:
    #   Sends the given message to the Tello.
    #   Returns the response string if the message was received.
    #   Returns None if the message failed.
    def __send(self, msg):
        self.log.append([msg, None])
        self.send_channel.sendto(msg.encode('utf-8'), self.tello_addr)
        # Response wait loop
        start = perf_counter()
        while self.log[-1][1] is None:
            if (perf_counter() - start) > self.MAX_TIME_OUT:
                self.log[-1][1] = "TIMED OUT"
                return None
        return self.log[-1][1]

    # Precond:
    #   mess is a string containing the message to send.
    #
    # Postcond:
    #   Sends the given message to the Tello.
    #   Does not wait for a response.
    #   Used (internally) only for sending the emergency signal (which is sent
    #       in triplicate.)
    def __send_nowait(self, msg):
        self.send_channel.sendto(msg.encode('utf-8'), self.tello_addr)
        return None

    # Precond:
    #   None.
    #
    # Postcond:
    #   Receives messages from the Tello and logs them.
    def __receive(self):
        while not self.stop:
            try:
                response, ip = self.send_channel.recvfrom(1024)
                response = response.decode('utf-8')
                self.log[-1][1] = response.strip()
            except OSError as exc:
                if not self.stop:
                    print("Caught exception socket.error : %s" % exc)